﻿//*************************************************************************************************
//【サンプルプログラム】『Kinectのデプスイメージから呼吸をモニターする』
//【名称】KinectVitalMonitor　【Version】0.01
//【作成者】私設研究所Neo-Tech-Lab.com  上田智章
//【作成日】2012年10月15日　【コメント追記】2012年10月15日
//*************************************************************************************************

using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Linq;
using System.Text;
using System.Windows.Forms;
using System.Diagnostics;             //【Debug】を使うための定義
using System.Drawing;                 //【Bitmap】【Color】を使うための定義
using System.Drawing.Imaging;         //【BitmapData】を使うための定義
using System.Runtime.InteropServices; //【Marshal.Copy】を使うための定義
using Microsoft.Kinect;               //【参照設定】プロジェクト(P)⇒参照の追加(F)⇒.NETタブ⇒Microsoft.Kinectを選択

namespace KinectVitalMonitor2
{
    public partial class Form1 : Form
    {
        //【KINECT関連】
        //　関数内でダイナミックに変数を確保すると動作速度が遅くなるので、ここにまとめて変数を確保しています。
        private KinectSensor KS_KinectSensor;   //【KINECT】
        private DepthImageFrame KS_DepthFrame;  //【KINECTのDepthImageFrame】
        private short[] KS_DepthData;           //【KINECTのDepthImageData】16bit/画素(MSB13bit:Depth, LSB3bit:PlayerID) 
        private byte[] KS_DepthImageDataXRGB;   //【KINECTのDepthImageData】32bit/画素,X8R8G8B8表示形式
        private int KS_DepthFrameWidth;         //【KINECTのDepthFrameの横方向画素数】
        private int KS_DepthFrameHeight;        //【KINECTのDepthFrameの縦方向画素数】
        private Bitmap KS_bmp2;                 //【DepthImage表示用Bitmapオブジェクト】
        private BitmapData KS_data2;            //【DepthImage用BitmapData】
        private Rectangle KS_rect2;             //【DepthImage表示枠】
        private int KS_PlayerIndexBitmask;      //【DepthデータのPlayer情報】 0000 0000 0000 0111B = 7
        private int KS_PlayerIndexBitmaskWidth; //【DepthデータのPlayer情報】 3 bits
        private int KS_Lmin;                    //【深度測定可能最近距離】default:  800mm, near: 400mm 
        private int KS_Lmax;                    //【深度測定可能最遠距離】default: 4000mm, near:3000mm
        private bool KS_flagFlip = true;        //【左右反転フラグ】
        //【NTL拡張現実センサ関連】---------------------------------------------------------
        private NTL_ColorBar ColorBar = new NTL_ColorBar(); //【デプス値をカラーに変換する部品】
        private int NTL_Count = 0; //【計測回数】
        private float NTL_cSum;    //【領域内累積加算デプス値の現在値】
        private string NTL_Log=""; //【距離測定結果を残すための文字列】
        private const int NTL_nFIFO = 551;               //【波形表示の為のデータ格納配列サイズ】
        private float[] NTL_FIFO = new float[NTL_nFIFO]; //【波形表示の為のデータ格納用FIFO】
        private int NTL_ptr = 0;                         //【FIFOのポインタ】
        private const int NTL_nLPF = 30;                 //
        private const float NTL_kLPF = 1.0f / (float)NTL_nLPF;
        private float[] NTL_LPF = new float[NTL_nLPF];   //
        private int NTL_pLPF = 0;                        //
        private float NTL_LPFvalue = 0.0f;
        private System.Drawing.Point[] NTL_Points = new System.Drawing.Point[NTL_nFIFO];
        private int[] NTL_Rect = { 300, 220, 40, 40 };
        private float NTL_kRect;
        private int NTL_nBC = 30; //【2秒分】
        private NTL_EstimationParabola estBC = new NTL_EstimationParabola(); //【呼吸曲線推定】
        private System.Drawing.Point[] NTL_Parabola = new System.Drawing.Point[71];

        public Form1()
        {
            InitializeComponent();
            //***********************************************************
            //【Kinect検出】【共通】
            //***********************************************************
            //【利用可能なKINECTセンサを取得する】
            KS_KinectSensor = (from sensorToCheck
                               in KinectSensor.KinectSensors
                               where sensorToCheck.Status == KinectStatus.Connected
                               select sensorToCheck).FirstOrDefault();
            //【検出結果の表示】
            if (KS_KinectSensor == null) //【Kinect接続が確認できなかった場合】
            {
                pictureBox2.Image = imageList1.Images[1]; //【Kinect未検出の画像を表示】
                MessageBox.Show(
                   "接続可能なKINECTセンサが見当たりません。",
                   "KINECT for Windows Error Message",
                   MessageBoxButtons.OK
                );
                return;
            }
            else //【Kinect検出に成功】
            {
                pictureBox2.Image = imageList1.Images[0]; //【Kinectセンサ画像を表示】
            }

            //***********************************************************
            //【Kinectの初期化処理】
            //***********************************************************
            //【KINECTの設定】DepthStreamを許可
            KS_KinectSensor.DepthStream.Enable(DepthImageFormat.Resolution640x480Fps30);    //【DepthStream許可】
            // DepthImageFormatとしては他にResolution320x240Fps30とDepthImageFormat.Resolution80x60Fps30がある
            //【KINECT動作開始】
            KS_KinectSensor.Start(); //【KINECTスタート】

            //【KINECTデータ取得イベントの設定とデータバッファの確保】
            KS_KinectSensor.DepthFrameReady += new EventHandler<DepthImageFrameReadyEventArgs>(DepthImageReady); //【DepthImageReadyをcallするよう設定】
            KS_DepthData = new short[KS_KinectSensor.DepthStream.FramePixelDataLength];             //【深度情報データ１フレーム分を保存するメモリを確保】
            KS_DepthImageDataXRGB = new byte[KS_KinectSensor.DepthStream.FramePixelDataLength * 4]; //【深度情報データ表示用】

            KS_PlayerIndexBitmask = DepthImageFrame.PlayerIndexBitmask;           //【デプス値のPlayerIDのマスク 0000000000000111B】
            KS_PlayerIndexBitmaskWidth = DepthImageFrame.PlayerIndexBitmaskWidth; //【デプス値のPlayerIDのビット数 3】
            KS_DepthFrameWidth = KS_KinectSensor.DepthStream.FrameWidth;   //【DepthImageの横方向画素数】
            KS_DepthFrameHeight = KS_KinectSensor.DepthStream.FrameHeight; //【DepthImageの縦方向画素数】
            KS_bmp2 = new Bitmap(KS_DepthFrameWidth, KS_DepthFrameHeight, System.Drawing.Imaging.PixelFormat.Format32bppRgb); //【DepthImage表示用】
            KS_rect2 = new Rectangle(0, 0, KS_DepthFrameWidth, KS_DepthFrameHeight); //【DepthImage用枠】
      //    KS_KinectSensor.DepthStream.Range = DepthRange.Default; //★【Kinect for XBOX360】【Depth測定モードをDefaultに設定】
            KS_KinectSensor.DepthStream.Range = DepthRange.Near;    //【Kinect for Windows】【Depth測定モードをDefaultに設定】
            KS_CheckDistance();                                     //【設定結果を表示】Lmax, Lminを取得
            pictureBox1.SizeMode = PictureBoxSizeMode.StretchImage; //【デプス画像をpictureBox1全体に引き伸ばして表示する】

            //【Kinectヘッドの仰角を設定】設定範囲は±27°
            int Angle = -4;
            KS_KinectSensor.ElevationAngle = Angle; //【ベッドより上方から撮影するのでヘッドを下に向ける】★要調整
            textBox1.Text = Angle.ToString();

            //【深度情報をカラーに変換する為の準備処理】
            ColorBar.CreateColorPalletForAbsolute(KS_Lmax+1);         //【カラーパレットを作成する】
            pictureBox3.Image = ColorBar.CreateColorScale(256, 10); //【256×10画素サイズのカラーバーを作成して表示する】

            //【呼吸曲線推定】
            estBC.Initialize(NTL_nBC); //【過去1秒分】

            //【Form終了イベント】
            this.FormClosing += new FormClosingEventHandler(Form1_FormClosing);
            NTL_kRect = 1.0f / (float)(NTL_Rect[2] * NTL_Rect[3]);
            radioButton1.CheckedChanged += new EventHandler(DepthRangeChange);
            checkBox1.CheckedChanged += new EventHandler(checkBox1_CheckedChanged);
        }

        void Form1_FormClosing(object sender, FormClosingEventArgs e)
        {
            if (KS_KinectSensor != null)
            {
                KS_KinectSensor.Stop();    //【KINECTセンサを終了】★これを実行しないとレーザーが消えない
                KS_KinectSensor.Dispose(); //【KINECTセンサのオブジェクトを解放する】
            }
        }

        //***************************************************************
        //【距離モード変更】Defaultモード(800mm～4000mm) Nearモード(400m～3000mm)
        //***************************************************************
        void DepthRangeChange(object sender, EventArgs e)
        {
            if (radioButton1.Checked) { KS_KinectSensor.DepthStream.Range = DepthRange.Default; label8.Text = "Default"; }
            else { KS_KinectSensor.DepthStream.Range = DepthRange.Near; label8.Text = "Near"; }
            KS_CheckDistance();
        }

        //***************************************************************
        //【CheckDistance】KINECTセンサの検出可能距離現在値を調べる
        //***************************************************************
        private void KS_CheckDistance()
        {
            KS_Lmax = KS_KinectSensor.DepthStream.MaxDepth;
            KS_Lmin = KS_KinectSensor.DepthStream.MinDepth;
            label1.Text = "Min=" + KS_Lmin.ToString() + "mm";
            label2.Text = "Max=" + KS_Lmax.ToString() + "mm";
        }

        //***************************************************************
        //【上下角変更ボタン】KINECTの上下角を変更。±27°の範囲内。時間を置かないとエラーが発生する
        //***************************************************************
        private void button2_Click(object sender, EventArgs e)
        {
            Int32 angle = int.Parse(textBox1.Text); //【設定角度を読み込む】±27°
            if (angle <= KS_KinectSensor.MaxElevationAngle && angle >= KS_KinectSensor.MinElevationAngle)
            {
                KS_KinectSensor.ElevationAngle = angle; //【上下角の変更】
            }
        }

        //***************************************************************
        //【カラー画像の左右反転指定フラグ】
        //***************************************************************
        private void checkBox1_CheckedChanged(object sender, EventArgs e)
        {
            if (checkBox1.Checked) { KS_flagFlip = true; } else { KS_flagFlip = false; }
        }

        //***************************************************************
        //【DepthImageの取得イベント】
        //***************************************************************
        private void DepthImageReady(object sender, DepthImageFrameReadyEventArgs e)
        {
            int i;
            int ix0 = NTL_Rect[0]; //【距離測定を行う領域のx座標】
            int iy0 = NTL_Rect[1]; //【距離測定を行う領域のy座標】
            int iLx = NTL_Rect[2]; //【距離測定を行う領域のx方向の幅】
            int iLy = NTL_Rect[3]; //【距離測定を行う領域のy方向の幅】
            float Dx, Dy;
            float v = 0;
            float maxValue, minValue;
            float t, O;
            using (KS_DepthFrame = e.OpenDepthImageFrame())
            {
                if (KS_DepthFrame != null)
                {
                    KS_DepthFrame.CopyPixelDataTo(KS_DepthData); //【KINECTのDepth情報を配列DepthDataにコピー】
                    KS_ConvertDepthToColor();                    //【深度情報をカラーバーの色を使って表示形式をX8R8G8B8形式に変換】
                    KS_data2 = KS_bmp2.LockBits(KS_rect2, ImageLockMode.WriteOnly, System.Drawing.Imaging.PixelFormat.Format32bppRgb); //【Lockする】
                    Marshal.Copy(KS_DepthImageDataXRGB, 0, KS_data2.Scan0, KS_DepthImageDataXRGB.Length); //【UnmanagedMemoryのBitmapDataに配列depthDataをコピーする】
                    KS_bmp2.UnlockBits(KS_data2);                                           //【Lock解除】
                    if (KS_flagFlip) { KS_bmp2.RotateFlip(RotateFlipType.Rotate180FlipY); } //【左右反転】
                    pictureBox1.Image = KS_bmp2;                                            //【pictureBox1に画像表示】

                    NTL_cSum = GetRectangleDistance(ix0, iy0, iLx, iLy, NTL_kRect); //【距離測定】指定領域内の加算平均デプス値を求める

                    Graphics g = Graphics.FromImage(pictureBox1.Image);        //【pixtureBox1内に測定値を重畳表示】
                    Color lc = Color.Azure;                                    //【領域位置表示色の定義】
                    if (KS_flagFlip) { ix0 = KS_DepthFrameWidth - ix0 - iLx; } //【左右反転表示のための処理】
                    g.DrawRectangle(new Pen(lc, 3.0f), ix0, iy0, iLx, iLy);    //【観測領域】
                    // g.DrawRectangle(new Pen(lc, 3.0f), ix1, iy1, iLx, iLy); //【観測領域2】
                    g.DrawLine(new Pen(lc, 3.0f), 320, 0, 320, 480);           //【画面中央縦線】
                    g.DrawLine(new Pen(lc, 3.0f), 0, 240, 640, 240);           //【画面中央横線】

                    if (NTL_Count == 0) //【初回はFIFOを初期化する】
                    {
                        NTL_ptr = 0; NTL_pLPF = 0; //【ポインタ初期化】
                        NTL_LPFvalue = NTL_cSum * (float)NTL_nLPF; //【累積加算値】
                        for (i = 0; i < NTL_nLPF; i++) { NTL_LPF[i] = NTL_cSum; }
                        for (i = 0; i < NTL_nFIFO; i++) { NTL_FIFO[i] = NTL_LPFvalue * NTL_kLPF; }
                    }

                    //【LPF】
                    NTL_LPFvalue += NTL_cSum - NTL_LPF[NTL_pLPF];
                    NTL_LPF[NTL_pLPF] = NTL_cSum;
                    NTL_pLPF++; if (NTL_pLPF == NTL_nLPF) { NTL_pLPF = 0; } //【FIFO管理ポインタを更新する】
                    //【FIFO】
                    NTL_FIFO[NTL_ptr] = (NTL_LPFvalue * NTL_kLPF);        //【波形表示用の測定値を保存する】
                    NTL_ptr++; if (NTL_ptr == NTL_nFIFO) { NTL_ptr = 0; }   //【FIFO管理ポインタを更新する】

                    v = NTL_LPFvalue * NTL_kLPF; // *0.001f;
                    estBC.Estimation(v); //【呼吸曲線推定処理】


                    //【FIFO内の最大値と最小値を求める】
                    minValue = 8000.0f;
                    maxValue = 0.0f;
                    for (i = 0; i < NTL_nFIFO; i++)
                    {
                        if (NTL_FIFO[i] > maxValue) maxValue = NTL_FIFO[i];
                        if (NTL_FIFO[i] < minValue) minValue = NTL_FIFO[i];
                    }
                    if (maxValue < minValue + 2.0f) maxValue = minValue + 2.0f; //【最大±1mmで表示】

                    //【距離測定している矩形領域をデプス画像上に表示する】
                    g = Graphics.FromImage(pictureBox1.Image);
                    lc = Color.Red;
                    Dx = pictureBox1.Width / NTL_nFIFO;
                    Dy = pictureBox1.Height / (maxValue - minValue);
                    for (i = 0; i < NTL_nFIFO; i++)
                    {
                        NTL_Points[i].X = (int)(i*Dx);
                        NTL_Points[i].Y = (int)((NTL_FIFO[NTL_ptr] - minValue) * Dy);
                        NTL_ptr++; if (NTL_ptr == NTL_nFIFO) { NTL_ptr = 0; } //【FIFO管理ポインタを更新する】
                    }
                    g.DrawLines(new Pen(lc,5.0f),NTL_Points);

                    //【推定した呼吸曲線】
                    lc = Color.Violet;
                    for (i = -30; i <= 40; i++)
                    {
                        t = (float)i;
                        O = estBC.A * t * t + estBC.B * t + estBC.C;
                        NTL_Parabola[i + 30].X = (int)((i + NTL_nFIFO) * Dx);
                        NTL_Parabola[i + 30].Y = (int)((O - minValue) * Dy);
                    }
                    g.DrawLines(new Pen(lc, 5.0f), NTL_Parabola);


                    lc = Color.White;
                    g.DrawLine(new Pen(lc, 3.0f), 550, 0, 550, 480);           //【画面中央縦線】
                    g.Dispose();
                    Validate();

                    NTL_Count++; //【測定番号を+1する】
                }
                else
                {
                    //【DepthFrameがnullだったのは要求が間に合わなかったと考えられます】
                }
                if (NTL_Count < 3001) //【NTL_Countの0～3000まで距離測定を行う】
                {
                    NTL_Log += NTL_Count.ToString() + ", " + v.ToString() + ", " + estBC.A.ToString() + ", " + estBC.B.ToString() + ", " + estBC.C.ToString() + ", " + estBC.D.ToString() + "\r\n"; //【測定番号と測定値を記録】
                }
                if (NTL_Count == 3001) textBox2.Text = NTL_Log; //【測定終了時にテキストボックス内に結果を表示する】
            }
        }

        //***************************************************************
        //【GetRectangleDistance】指定矩形領域までの距離を取得
        //***************************************************************
        private float GetRectangleDistance(int x0, int y0, int w, int h, float keisu)
        {
            float sum = 0.0f;
            int i, j, k;
            for (j = y0; j < y0 + h; j++)
            {
                k = KS_DepthFrameWidth * j;
                for (i = x0; i < x0 + w; i++)
                {
                    sum += (KS_DepthData[i + k] >> KS_PlayerIndexBitmaskWidth); //【デプス値を累積加算】
                }
            }
            return sum * keisu;//【デプス値をサンプル数で割って加算平均値を得る】【Keisuは1.0/(w*h)】
        }

        //***************************************************************
        //【ConvertDepthToColor】16ビットのDepth値をカラーバーに対応する色に変更
        //***************************************************************
        private void KS_ConvertDepthToColor()
        {
            int i, j, d;  //【Nearモード】  400～3000mm ⇒ 0.0980769　2600色
            j = 0;
            for (i = 0; i < KS_DepthData.Length; i++)
            {
                d = KS_DepthData[i];                  //【距離情報】【上位13bitがDepth値】
                //      playerID = d & KS_PlayerIndexBitmask; //【下位3bitはplayerID】【未使用】
                d = ((d >> KS_PlayerIndexBitmaskWidth) & 0x1FFF);  //【realDepth】0mm～8191mm
                if (d < KS_Lmin) { d = KS_Lmin; } //【近すぎる】
                if (d > KS_Lmax) { d = KS_Lmax-1; } //【遠すぎる】
                d = d * 4;
                KS_DepthImageDataXRGB[j] = ColorBar.NTL_ColorElement[d]; //【青】B
                KS_DepthImageDataXRGB[j + 1] = ColorBar.NTL_ColorElement[d + 1]; //【緑】G
                KS_DepthImageDataXRGB[j + 2] = ColorBar.NTL_ColorElement[d + 2]; //【赤】R
                KS_DepthImageDataXRGB[j + 3] = ColorBar.NTL_ColorElement[d + 3]; ; //【α】
                j = j + 4;
            }
        }

        private void button3_Click(object sender, EventArgs e)
        {
            //【再計測の処理はNTL_Countを0にするだけで開始する】
            NTL_Count = 0;
            NTL_Log = "";
            textBox2.Text = "";
        }
    }
}
